@@ -4,15 +4,13 @@ const char * _messages_service =
 "	<interface name=\"com.canonical.indicator.messages.service\">\n"
 "\n"
 "<!-- Methods -->\n"
-"		<method name=\"Watch\">\n"
-"			<annotation name=\"org.freedesktop.DBus.GLib.Async\" value=\"true\" />\n"
-"		</method>\n"
 "		<method name=\"AttentionRequested\">\n"
 "			<arg type=\"b\" name=\"dot\" direction=\"out\" />\n"
 "		</method>\n"
 "		<method name=\"IconShown\">\n"
 "			<arg type=\"b\" name=\"hidden\" direction=\"out\" />\n"
 "		</method>\n"
+"		<method name=\"ClearAttention\" />\n"
 "\n"
 "<!-- Signals -->\n"
 "		<signal name=\"AttentionChanged\">\n"
